import typing as t
import pyarrow as pa
import numpy as np
import pyarrow.compute as pc
import warnings

from sarus_data_spec.arrow.array import convert_record_batch
from sarus_data_spec.constants import (
    DATA,
    DATASET_SLUGNAME,
    MULTIPLICITY,
)
from sarus_data_spec.manager.ops.processor.standard.standard_op import (  # noqa: E501
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
)

from sarus_data_spec.dataset import Dataset
from sarus_data_spec.path import Path, straight_path
from sarus_data_spec.marginals import marginals as marg_builder
from sarus_data_spec.multiplicity import multiplicity as multiplicity_builder
from sarus_data_spec.size import size as size_builder
from sarus_data_spec.bounds import bounds as bounds_builder
from sarus_data_spec.schema import schema
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st

try:
    import sqlalchemy as sa
except ModuleNotFoundError:
    warnings.warn("sqlalchemy not installed. No sampling on bigdata")


try:
    from sarus_data_spec.manager.ops.sql_utils.bigdata import (
        path_to_quoted_string,
        sqlalchemy_query_to_string,
    )
    from sarus_data_spec.manager.ops.sql_utils.schema_translations import (
        async_sa_metadata_from_dataset,
    )
except ModuleNotFoundError:
    warnings.warn("sql utils not installed.")


class SelectTableStaticChecker(StandardDatasetStaticChecker):
    async def schema(self) -> st.Schema:
        parent_schema = await self.parent_schema()
        path = Path(self.dataset.transform().protobuf().spec.select_table.path)
        sub_types = parent_schema.data_type().sub_types(path)
        assert len(sub_types) == 1
        new_type = sub_types[0]
        # TODO: update foreign_keys/primary_keys in the type
        previous_fields = parent_schema.type().children()

        try:
            parent_stats = (await self.parent_marginals()).statistics()
            path = Path(
                self.dataset.transform().protobuf().spec.select_table.path
            )
            tab_multiplicity = parent_stats.nodes_statistics(path)[
                0
            ].multiplicity()
            properties = {MULTIPLICITY: str(tab_multiplicity)}
        except Exception:
            properties = {}
        if DATA in previous_fields.keys():
            previous_fields[DATA] = new_type
            new_type = sdt.Struct(fields=previous_fields)
        return schema(
            self.dataset,
            schema_type=new_type,
            privacy_unit_tracking_paths=None,
            name=self.dataset.properties().get(DATASET_SLUGNAME, None),
            properties=properties,
        )


class SelectTable(StandardDatasetImplementation):
    """Computes schema and arrow
    batches for a dataspec transformed by
    a select_table transform
    """

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        previous_ds = t.cast(Dataset, self.parent())
        path = Path(self.dataset.transform().protobuf().spec.select_table.path)
        parent_schema = await self.parent_schema()

        async def select_table_func(
            batch: pa.RecordBatch,
        ) -> pa.Array:
            array = convert_record_batch(
                record_batch=batch, _type=parent_schema.type()
            )

            if DATA in parent_schema.type().children():
                old_arrays = array.flatten()
                idx_data = array.type.get_field_index(DATA)
                array = old_arrays[idx_data]
                updated_array, filter_indices = select_table(
                    _type=parent_schema.data_type(),
                    array=array,
                    path=path,
                )
                old_arrays[idx_data] = updated_array

                new_struct = pa.StructArray.from_arrays(
                    old_arrays,
                    names=list(parent_schema.type().children().keys()),
                )
                return new_struct.filter(filter_indices)

            updated_array, filter_indices = select_table(
                array=array, path=path, _type=parent_schema.data_type()
            )
            if isinstance(updated_array, pa.StructArray):
                return updated_array.filter(filter_indices)
            else:
                old_arrays[idx_data] = updated_array

                new_struct = pa.StructArray.from_arrays(
                    old_arrays,
                    names=[path.to_strings_list()[0][-1]],
                )
                return new_struct.filter(filter_indices)

        return await self.ensure_batch_correct(
            async_iterator=await previous_ds.async_to_arrow(
                batch_size=batch_size
            ),
            batch_size=batch_size,
            func_to_apply=select_table_func,
        )

    async def sql_implementation(
        self,
    ) -> t.Optional[t.Dict[t.Tuple[str, ...], str]]:
        sqlalchemy_metadata = await async_sa_metadata_from_dataset(
            t.cast(Dataset, self.parent())
        )
        queries: t.Dict[t.Tuple[str, ...], str] = {}
        table_path = Path(
            self.dataset.transform().protobuf().spec.select_table.path
        )
        curr_path = table_path.to_strings_list()[0]
        full_tablename = curr_path[1:] if curr_path[0] == DATA else curr_path
        sa_table_name = path_to_quoted_string(straight_path(full_tablename))
        sa_table = sqlalchemy_metadata.tables[sa_table_name]
        table_query = sa.select(sa_table)

        queries[()] = sqlalchemy_query_to_string(table_query)
        return queries

    async def size(self) -> st.Size:
        sizes = await self.parent_size()
        path = Path(self.dataset.transform().protobuf().spec.select_table.path)
        new_stats = sizes.statistics().nodes_statistics(path)
        assert len(new_stats) == 1
        return size_builder(dataset=self.dataset, statistics=new_stats[0])

    async def multiplicity(self) -> st.Multiplicity:
        multiplicities = await self.parent_multiplicity()
        path = Path(self.dataset.transform().protobuf().spec.select_table.path)
        new_stats = multiplicities.statistics().nodes_statistics(path)
        assert len(new_stats) == 1
        return multiplicity_builder(
            dataset=self.dataset, statistics=new_stats[0]
        )

    async def bounds(self) -> st.Bounds:
        bounds = await self.parent_bounds()
        path = Path(self.dataset.transform().protobuf().spec.select_table.path)
        new_stats = bounds.statistics().nodes_statistics(path)
        assert len(new_stats) == 1
        return bounds_builder(dataset=self.dataset, statistics=new_stats[0])

    async def marginals(self) -> st.Marginals:
        marginals = await self.parent_marginals()
        path = Path(self.dataset.transform().protobuf().spec.select_table.path)
        new_stats = marginals.statistics().nodes_statistics(path)
        assert len(new_stats) == 1
        return marg_builder(dataset=self.dataset, statistics=new_stats[0])


def select_table(
    array: pa.Array, path: st.Path, _type: st.Type, is_optional: bool = False
) -> t.Tuple[pa.Array, pa.Array]:
    """Visitor selecting columns based on the type."""

    class TableSelector(sdt.TypeVisitor_):
        batch_array: pa.Array = array
        filter_indices: pa.Array

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if len(path.sub_paths()) == 0:
                self.batch_array = array
                self.filter_indices = np.full_like(array, True)
            else:
                sub_path = path.sub_paths()[0]
                idx = array.type.get_field_index(sub_path.label())

                initial_field_selected = self.batch_array.field(
                    "field_selected"
                ).to_numpy(zero_copy_only=False)

                self.batch_array, filter_idx = select_table(
                    array=array.flatten()[idx],
                    path=sub_path,
                    _type=fields[sub_path.label()],
                )

                field_selected = pa.array(
                    initial_field_selected == sub_path.label(), type=pa.bool_()
                )
                result = pc.and_(field_selected, filter_idx)
                self.filter_indices = result

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.filter_indices = pa.array(
                [True] * len(self.batch_array), type=pa.bool_()
            )

    visitor = TableSelector()
    _type.accept(visitor)
    return visitor.batch_array, visitor.filter_indices



from typing import Any, Optional

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from sklearn import linear_model
except ModuleNotFoundError:
    pass  # error message in typing.py


class sk_linear_regression(ExternalOpImplementation):
    _transform_id = "sklearn.SK_LINEAR_REGRESSION"
    _signature = SarusSignature(
        SarusParameter(
            name="fit_intercept",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="copy_X",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="positive",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return linear_model.LinearRegression(**kwargs)


class sk_logistic_regression(ExternalOpImplementation):
    _transform_id = "sklearn.SK_LOGISTIC_REGRESSION"
    _signature = SarusSignature(
        SarusParameter(
            name="penalty",
            annotation=str,
            default="l2",
        ),
        SarusParameter(
            name="dual",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="tol",
            annotation=float,
            default=0.0001,
        ),
        SarusParameter(
            name="C",
            annotation=float,
            default=1.0,
        ),
        SarusParameter(
            name="fit_intercept",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="intercept_scaling",
            annotation=int,
            default=1,
        ),
        SarusParameter(
            name="class_weight",
            annotation=Optional[dict],
            default=None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="solver",
            annotation=str,
            default="lbfgs",
        ),
        SarusParameter(
            name="max_iter",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="multi_class",
            annotation=str,
            default="auto",
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="warm_start",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="l1_ratio",
            annotation=Optional[float],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return linear_model.LogisticRegression(**kwargs)



import typing as t

import pyarrow as pa

import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st

INTBASE_TO_ARROW = {
    st.IntegerBase.INT64: pa.int64(),
    st.IntegerBase.INT32: pa.int32(),
    st.IntegerBase.INT16: pa.int16(),
    st.IntegerBase.INT8: pa.int8(),
    st.IntegerBase.UINT64: pa.uint64(),
    st.IntegerBase.UINT32: pa.uint32(),
    st.IntegerBase.UINT16: pa.uint16(),
    st.IntegerBase.UINT8: pa.uint8(),
}

IDBASE_TO_ARROW = {
    st.IdBase.INT64: pa.int64(),
    st.IdBase.INT32: pa.int32(),
    st.IdBase.INT16: pa.int16(),
    st.IdBase.INT8: pa.int8(),
    st.IdBase.STRING: pa.large_string(),
    st.IdBase.BYTES: pa.binary(),
}

FLOATBASE_TO_ARROW = {
    st.FloatBase.FLOAT64: pa.float64(),
    st.FloatBase.FLOAT32: pa.float32(),
    st.FloatBase.FLOAT16: pa.float16(),
}

DATETIMEBASE_TO_ARROW = {
    st.DatetimeBase.INT64_NS: pa.timestamp("ns"),
    st.DatetimeBase.INT64_MS: pa.timestamp("ms"),
    st.DatetimeBase.STRING: pa.large_string(),
}

DATEBASE_TO_ARROW = {
    st.DateBase.INT32: pa.date32(),
    st.DateBase.STRING: pa.large_string(),
}

TIMEBASE_TO_ARROW = {
    st.TimeBase.INT64_US: pa.time64("us"),
    st.TimeBase.INT32_MS: pa.time32("ms"),
    st.TimeBase.STRING: pa.large_string(),
}


def to_arrow(
    _type: st.Type,
    nullable: bool = True,
) -> pa.DataType:
    """Visitor that maps sarus types to pa types
    See https://arrow.apache.org/docs/python/api/datatypes.html
    """

    class ToArrow(st.TypeVisitor):
        pa_type: pa.DataType

        def Null(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            self.pa_type = pa.null()

        def Unit(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            self.pa_type = pa.null()

        def Boolean(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            self.pa_type = pa.bool_()

        def Id(
            self,
            unique: bool,
            reference: t.Optional[st.Path] = None,
            base: t.Optional[st.IdBase] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if base is not None:
                self.pa_type = IDBASE_TO_ARROW[base]

        def Integer(
            self,
            min: int,
            max: int,
            base: st.IntegerBase,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.pa_type = INTBASE_TO_ARROW[base]

        def Enum(
            self,
            name: str,
            name_values: t.Sequence[t.Tuple[str, int]],
            ordered: bool,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.pa_type = pa.large_string()

        def Float(
            self,
            min: float,
            max: float,
            base: st.FloatBase,
            possible_values: t.Iterable[float],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.pa_type = FLOATBASE_TO_ARROW[base]

        def Text(
            self,
            encoding: str,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.pa_type = pa.large_string()

        def Bytes(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            self.pa_type = pa.binary()

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.pa_type = pa.struct(
                [
                    pa.field(
                        name=name,
                        type=to_arrow(field),
                        nullable=field.protobuf().HasField("optional")
                        or field.protobuf().HasField("unit"),
                    )
                    for name, field in fields.items()
                ]
            )

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.pa_type = pa.struct(
                [
                    pa.field(
                        name=name,
                        type=to_arrow(field),
                        nullable=True,
                    )
                    for name, field in fields.items()
                ]
                + [
                    pa.field(
                        name="field_selected",
                        type=pa.large_string(),
                        nullable=False,
                    )
                ]
            )

        def Optional(
            self,
            type: st.Type,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.pa_type = to_arrow(type, nullable=True)

        def List(
            self,
            type: st.Type,
            max_size: int,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.pa_type = pa.list_(
                to_arrow(type), -1
            )  # return variable length list

        def Array(
            self,
            type: st.Type,
            shape: t.Tuple[int, ...],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Datetime(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DatetimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.pa_type = DATETIMEBASE_TO_ARROW[base]

        def Time(
            self,
            format: str,
            min: str,
            max: str,
            base: st.TimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.pa_type = TIMEBASE_TO_ARROW[base]

        def Date(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DateBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.pa_type = DATEBASE_TO_ARROW[base]

        def Duration(
            self,
            unit: str,
            min: int,
            max: int,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            self.pa_type = pa.duration(unit)

        def Constrained(
            self,
            type: st.Type,
            constraint: st.Predicate,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Hypothesis(
            self,
            *types: t.Tuple[st.Type, float],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

    visitor = ToArrow()
    _type.accept(visitor)
    return visitor.pa_type


def from_arrow(type: pa.DataType) -> sdt.Type:
    # Integers
    if pa.types.is_int8(type):
        return sdt.Integer(base=st.IntegerBase.INT8)
    if pa.types.is_int16(type):
        return sdt.Integer(base=st.IntegerBase.INT16)
    if pa.types.is_int32(type):
        return sdt.Integer(base=st.IntegerBase.INT32)
    if pa.types.is_int64(type):
        return sdt.Integer(base=st.IntegerBase.INT64)
    if pa.types.is_uint8(type):
        return sdt.Integer(base=st.IntegerBase.UINT8)
    if pa.types.is_uint16(type):
        return sdt.Integer(base=st.IntegerBase.UINT16)
    if pa.types.is_uint32(type):
        return sdt.Integer(base=st.IntegerBase.UINT32)
    if pa.types.is_uint64(type):
        return sdt.Integer(base=st.IntegerBase.UINT64)

    # Floats
    if pa.types.is_float16(type):
        return sdt.Float(base=st.FloatBase.FLOAT16)
    if pa.types.is_float32(type):
        return sdt.Float(base=st.FloatBase.FLOAT32)
    if pa.types.is_float64(type):
        return sdt.Float(base=st.FloatBase.FLOAT64)
    if pa.types.is_string(type):
        return sdt.Text()
    if pa.types.is_large_string(type):
        return sdt.Text()
    if pa.types.is_boolean(type):
        return sdt.Boolean()

    # Temporal
    if pa.types.is_temporal(type):
        # Return True if value is an instance of date, time,
        # timestamp or duration.
        if pa.types.is_timestamp(type):
            return sdt.Datetime(base=st.DatetimeBase.INT64_NS)
            # TODO: when we will support different bases for datetime
            # we need to remove the asserts in the Datetime builder and
            # the error rise in the check_visitor in user_settings
        if pa.types.is_time(type):
            if type.unit in ["ns", "us"]:
                return sdt.Time(base=st.TimeBase.INT64_US)
            else:
                return sdt.Time(base=st.TimeBase.INT32_MS)

        if pa.types.is_date(type):
            return sdt.Date()
        else:
            # It is a duration
            if type.unit in ["s", "ms", "us"]:
                return sdt.Duration(unit=type.unit)
            else:
                raise ValueError(
                    "Duration type with nanosecond resolution not supported"
                )
    if pa.types.is_null(type):
        return sdt.Unit()
    if pa.types.is_struct(type):
        struct_type = t.cast(pa.StructType, type)
        return sdt.Struct(
            {
                struct_type.field(i).name: type_from_arrow(
                    struct_type.field(i).type, nullable=False
                )
                for i in range(struct_type.num_fields)
            }
        )
    if pa.types.is_list(type):
        list_type = t.cast(pa.ListType, type)
        return sdt.List(
            type=type_from_arrow(list_type.value_type, nullable=False)
        )
    raise NotImplementedError(f"Type {type} not implemented")


def type_from_arrow(
    arrow_type: pa.DataType,
    nullable: bool,
) -> st.Type:
    if nullable and not (pa.types.is_null(arrow_type)):
        return sdt.Optional(type=from_arrow(arrow_type))
    return from_arrow(arrow_type)



from __future__ import annotations

import typing as t

import numpy as np

from sarus_data_spec.base import Base
from sarus_data_spec.constants import ARRAY_VALUES, LIST_VALUES, OPTIONAL_VALUE
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class Statistics(Base[sp.Statistics]):
    """A python class to describe statistics"""

    def prototype(self) -> t.Type[sp.Statistics]:
        """Return the type of the underlying protobuf."""
        return sp.Statistics

    def name(self) -> str:
        """Return the statistics name"""
        return self.protobuf().name

    def distribution(self) -> st.Distribution:
        return Distribution(
            getattr(
                self.protobuf(),
                t.cast(str, self.protobuf().WhichOneof("statistics")),
            ).distribution
        )

    def size(self) -> int:
        return t.cast(
            int,
            getattr(
                self.protobuf(),
                t.cast(str, self.protobuf().WhichOneof("statistics")),
            ).size,
        )

    def multiplicity(self) -> float:
        return t.cast(
            float,
            getattr(
                self.protobuf(),
                t.cast(str, self.protobuf().WhichOneof("statistics")),
            ).multiplicity,
        )

    def accept(self, visitor: st.StatisticsVisitor) -> None:
        dispatch: t.Callable[[], None] = {
            "null": lambda: visitor.Null(
                self._protobuf.null.size, self._protobuf.null.multiplicity
            ),
            "unit": lambda: visitor.Unit(
                self._protobuf.unit.size, self._protobuf.unit.multiplicity
            ),
            "boolean": lambda: visitor.Boolean(
                size=self._protobuf.boolean.size,
                multiplicity=self._protobuf.boolean.multiplicity,
                probabilities=[
                    element.probability
                    for element in self._protobuf.boolean.distribution.boolean.points  # noqa: E501
                ],
                names=[
                    element.name
                    for element in self._protobuf.boolean.distribution.boolean.points  # noqa: E501
                ],
                values=[
                    element.value
                    for element in self._protobuf.boolean.distribution.boolean.points  # noqa: E501
                ],
            ),
            "integer": lambda: visitor.Integer(
                size=self._protobuf.integer.size,
                multiplicity=self._protobuf.integer.multiplicity,
                min_value=self._protobuf.integer.distribution.integer.min,
                max_value=self._protobuf.integer.distribution.integer.max,
                probabilities=[
                    element.probability
                    for element in self._protobuf.integer.distribution.integer.points  # noqa: E501
                ],
                values=[
                    element.value
                    for element in self._protobuf.integer.distribution.integer.points  # noqa: E501
                ],
            ),
            "id": lambda: visitor.Id(
                self._protobuf.id.size, self._protobuf.id.multiplicity
            ),
            "enum": lambda: visitor.Enum(
                size=self._protobuf.enum.size,
                multiplicity=self._protobuf.enum.multiplicity,
                probabilities=[
                    element.probability
                    for element in self._protobuf.enum.distribution.enum.points  # noqa: E501
                ],
                names=[
                    element.name
                    for element in self._protobuf.enum.distribution.enum.points  # noqa: E501
                ],
                values=[
                    element.value
                    for element in self._protobuf.enum.distribution.enum.points  # noqa: E501
                ],
            ),
            "float": lambda: visitor.Float(
                size=self._protobuf.float.size,
                multiplicity=self._protobuf.float.multiplicity,
                min_value=self._protobuf.float.distribution.double.min,
                max_value=self._protobuf.float.distribution.double.max,
                probabilities=[
                    element.probability
                    for element in self._protobuf.float.distribution.double.points  # noqa: E501
                ],
                values=[
                    element.value
                    for element in self._protobuf.float.distribution.double.points  # noqa: E501
                ],
            ),
            "text": lambda: visitor.Text(
                size=self._protobuf.text.size,
                multiplicity=self._protobuf.text.multiplicity,
                min_value=self._protobuf.text.distribution.integer.min,
                max_value=self._protobuf.text.distribution.integer.max,
                probabilities=[
                    element.probability
                    for element in self._protobuf.text.distribution.integer.points  # noqa: E501
                ],
                values=[
                    element.value
                    for element in self._protobuf.text.distribution.integer.points  # noqa: E501
                ],
                example=self._protobuf.text.example,
            ),
            "bytes": lambda: visitor.Bytes(
                self._protobuf.bytes.size, self._protobuf.bytes.multiplicity
            ),
            "struct": lambda: visitor.Struct(
                {
                    field.name: Statistics(field.statistics)
                    for field in self._protobuf.struct.fields
                },
                self._protobuf.struct.size,
                name=self._protobuf.name,
                multiplicity=self._protobuf.struct.multiplicity,
                properties=self._protobuf.properties,
            ),
            "union": lambda: visitor.Union(
                {
                    field.name: Statistics(field.statistics)
                    for field in self._protobuf.union.fields
                },
                size=self._protobuf.union.size,
                name=self._protobuf.name,
                multiplicity=self._protobuf.union.multiplicity,
                properties=self._protobuf.properties,
            ),
            "optional": lambda: visitor.Optional(
                Statistics(self._protobuf.optional.statistics),
                size=self._protobuf.optional.size,
                multiplicity=self._protobuf.optional.multiplicity,
            ),
            "list": lambda: visitor.List(
                statistics=Statistics(self._protobuf.list.statistics),
                size=self._protobuf.list.size,
                multiplicity=self._protobuf.list.multiplicity,
                min_value=self._protobuf.list.distribution.integer.min,
                max_value=self._protobuf.list.distribution.integer.max,
                probabilities=[
                    element.probability
                    for element in self._protobuf.list.distribution.integer.points  # noqa: E501
                ],
                values=[
                    element.value
                    for element in self._protobuf.list.distribution.integer.points  # noqa: E501
                ],
            ),
            "array": lambda: visitor.Array(
                statistics=Statistics(self._protobuf.array.statistics),
                size=self._protobuf.array.size,
                multiplicity=self._protobuf.array.multiplicity,
                min_values=[
                    distribution.double.min
                    for distribution in self._protobuf.array.distributions
                ],
                max_values=[
                    distribution.double.max
                    for distribution in self._protobuf.array.distributions
                ],
                probabilities=[
                    [
                        element.probability
                        for element in distribution.double.points
                    ]
                    for distribution in self._protobuf.array.distributions
                ],
                values=[
                    [element.value for element in distribution.double.points]
                    for distribution in self._protobuf.array.distributions
                ],
            ),
            "datetime": lambda: visitor.Datetime(
                size=self._protobuf.datetime.size,
                multiplicity=self._protobuf.datetime.multiplicity,
                min_value=self._protobuf.datetime.distribution.integer.min,
                max_value=self._protobuf.datetime.distribution.integer.max,
                probabilities=[
                    element.probability
                    for element in self._protobuf.datetime.distribution.integer.points  # noqa: E501
                ],
                values=[
                    element.value
                    for element in self._protobuf.datetime.distribution.integer.points  # noqa: E501
                ],
            ),
            "date": lambda: visitor.Date(
                size=self._protobuf.date.size,
                multiplicity=self._protobuf.date.multiplicity,
                min_value=self._protobuf.date.distribution.integer.min,
                max_value=self._protobuf.date.distribution.integer.max,
                probabilities=[
                    element.probability
                    for element in self._protobuf.date.distribution.integer.points  # noqa: E501
                ],
                values=[
                    element.value
                    for element in self._protobuf.date.distribution.integer.points  # noqa: E501
                ],
            ),
            "time": lambda: visitor.Time(
                size=self._protobuf.time.size,
                multiplicity=self._protobuf.time.multiplicity,
                min_value=self._protobuf.time.distribution.integer.min,
                max_value=self._protobuf.time.distribution.integer.max,
                probabilities=[
                    element.probability
                    for element in self._protobuf.time.distribution.integer.points  # noqa: E501
                ],
                values=[
                    element.value
                    for element in self._protobuf.time.distribution.integer.points  # noqa: E501
                ],
            ),
            "duration": lambda: visitor.Duration(
                size=self._protobuf.duration.size,
                multiplicity=self._protobuf.duration.multiplicity,
                min_value=self._protobuf.duration.distribution.integer.min,
                max_value=self._protobuf.duration.distribution.integer.max,
                probabilities=[
                    element.probability
                    for element in self._protobuf.duration.distribution.integer.points  # noqa: E501
                ],
                values=[
                    element.value
                    for element in self._protobuf.duration.distribution.integer.points  # noqa: E501
                ],
            ),
            "constrained": lambda: visitor.Constrained(
                Statistics(self._protobuf.constrained.statistics),
                self._protobuf.constrained.size,
                self._protobuf.constrained.multiplicity,
            ),
            None: lambda: None,
        }[self._protobuf.WhichOneof("statistics")]
        dispatch()

    def nodes_statistics(self, path: st.Path) -> t.List[st.Statistics]:
        """Returns the List of each statistics corresponding at the leaves of
        path"""

        class Select(st.StatisticsVisitor):
            def __init__(self, statistics: st.Statistics):
                self.result = [statistics]

            def Struct(
                self,
                fields: t.Mapping[str, st.Statistics],
                size: int,
                multiplicity: float,
                name: t.Optional[str] = None,
                properties: t.Optional[t.Mapping[str, str]] = None,
            ) -> None:
                result = []
                for sub_path in path.sub_paths():
                    result.extend(
                        fields[sub_path.label()].nodes_statistics(sub_path)
                    )
                if len(result) > 0:
                    self.result = result
                    # otherwise struct is empty and it is a terminal node

            def Union(
                self,
                fields: t.Mapping[str, st.Statistics],
                size: int,
                multiplicity: float,
                name: t.Optional[str] = None,
                properties: t.Optional[t.Mapping[str, str]] = None,
            ) -> None:
                result = []
                for sub_path in path.sub_paths():
                    result.extend(
                        fields[sub_path.label()].nodes_statistics(sub_path)
                    )
                if len(result) > 0:
                    self.result = result  # otherwise union is empty and it is a terminal node  # noqa: E501

            def Optional(
                self,
                statistics: st.Statistics,
                size: int,
                multiplicity: float,
            ) -> None:
                result = []
                if len(path.sub_paths()) == 1:
                    result.extend(
                        statistics.nodes_statistics(path.sub_paths()[0])
                    )
                    self.result = result

            def List(
                self,
                statistics: st.Statistics,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                name: str = "List",
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                result = []
                if len(path.sub_paths()) == 1:
                    result.extend(
                        statistics.nodes_statistics(path.sub_paths()[0])
                    )
                    self.result = result

            def Array(
                self,
                statistics: st.Statistics,
                size: int,
                multiplicity: float,
                min_values: t.Optional[t.List[float]] = None,
                max_values: t.Optional[t.List[float]] = None,
                name: str = "Array",
                probabilities: t.Optional[t.List[t.List[float]]] = None,
                values: t.Optional[t.List[t.List[float]]] = None,
            ) -> None:
                result = []
                if len(path.sub_paths()) == 1:
                    result.extend(
                        statistics.nodes_statistics(path.sub_paths()[0])
                    )
                    self.result = result

            def Constrained(
                self,
                statistics: st.Statistics,
                size: int,
                multiplicity: float,
            ) -> None:
                result = []
                if len(path.sub_paths()) == 1:
                    result.extend(
                        statistics.nodes_statistics(path.sub_paths()[0])
                    )
                    self.result = result

            def Bytes(self, size: int, multiplicity: float) -> None:
                pass

            def Date(
                self,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

            def Time(
                self,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

            def Datetime(
                self,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

            def Duration(
                self,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

            def Null(self, size: int, multiplicity: float) -> None:
                pass

            def Unit(self, size: int, multiplicity: float) -> None:
                pass

            def Boolean(
                self,
                size: int,
                multiplicity: float,
                probabilities: t.Optional[t.List[float]] = None,
                names: t.Optional[t.List[bool]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

            def Id(self, size: int, multiplicity: float) -> None:
                pass

            def Integer(
                self,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

            def Enum(
                self,
                size: int,
                multiplicity: float,
                probabilities: t.Optional[t.List[float]] = None,
                names: t.Optional[t.List[str]] = None,
                values: t.Optional[t.List[float]] = None,
                name: str = "Enum",
            ) -> None:
                pass

            def Float(
                self,
                size: int,
                multiplicity: float,
                min_value: float,
                max_value: float,
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[float]] = None,
            ) -> None:
                pass

            def Text(
                self,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                example: str = "",
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

        visitor = Select(statistics=self)
        self.accept(visitor)
        return visitor.result

    def children(self: st.Statistics) -> t.Dict[str, st.Statistics]:
        """Returns the children contained in the type tree structure"""

        class GetChildren(st.StatisticsVisitor):
            result: t.Dict[str, st.Statistics] = {}

            def __init__(
                self, properties: t.Optional[t.Mapping[str, str]] = None
            ):
                self.properties = properties

            def Struct(
                self,
                fields: t.Mapping[str, st.Statistics],
                size: int,
                multiplicity: float,
                name: t.Optional[str] = None,
                properties: t.Optional[t.Mapping[str, str]] = None,
            ) -> None:
                self.result = t.cast(t.Dict[str, st.Statistics], fields)

            def Union(
                self,
                fields: t.Mapping[str, st.Statistics],
                size: int,
                multiplicity: float,
                name: t.Optional[str] = None,
                properties: t.Optional[t.Mapping[str, str]] = None,
            ) -> None:
                self.result = t.cast(t.Dict[str, st.Statistics], fields)

            def Optional(
                self,
                statistics: st.Statistics,
                size: int,
                multiplicity: float,
            ) -> None:
                self.result = {OPTIONAL_VALUE: statistics}

            def List(
                self,
                statistics: st.Statistics,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                name: str = "List",
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                self.result = {LIST_VALUES: statistics}

            def Array(
                self,
                statistics: st.Statistics,
                size: int,
                multiplicity: float,
                min_values: t.Optional[t.List[float]] = None,
                max_values: t.Optional[t.List[float]] = None,
                name: str = "Array",
                probabilities: t.Optional[t.List[t.List[float]]] = None,
                values: t.Optional[t.List[t.List[float]]] = None,
            ) -> None:
                self.result = {ARRAY_VALUES: statistics}

            def Bytes(self, size: int, multiplicity: float) -> None:
                pass

            def Date(
                self,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

            def Time(
                self,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

            def Datetime(
                self,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

            def Duration(
                self,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

            def Null(self, size: int, multiplicity: float) -> None:
                pass

            def Unit(self, size: int, multiplicity: float) -> None:
                pass

            def Boolean(
                self,
                size: int,
                multiplicity: float,
                probabilities: t.Optional[t.List[float]] = None,
                names: t.Optional[t.List[bool]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

            def Id(self, size: int, multiplicity: float) -> None:
                pass

            def Integer(
                self,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

            def Enum(
                self,
                size: int,
                multiplicity: float,
                probabilities: t.Optional[t.List[float]] = None,
                names: t.Optional[t.List[str]] = None,
                values: t.Optional[t.List[float]] = None,
                name: str = "Enum",
            ) -> None:
                pass

            def Float(
                self,
                size: int,
                multiplicity: float,
                min_value: float,
                max_value: float,
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[float]] = None,
            ) -> None:
                pass

            def Text(
                self,
                size: int,
                multiplicity: float,
                min_value: int,
                max_value: int,
                example: str = "",
                probabilities: t.Optional[t.List[float]] = None,
                values: t.Optional[t.List[int]] = None,
            ) -> None:
                pass

            def Constrained(
                self, statistics: st.Statistics, size: int, multiplicity: float
            ) -> None:
                raise NotImplementedError

        visitor = GetChildren(properties=self.properties())
        self.accept(visitor)
        return visitor.result


class Distribution(Base[sp.Distribution]):
    """A python class to describe distributions"""

    def prototype(self) -> t.Type[sp.Distribution]:
        """Return the type of the underlying protobuf."""
        return sp.Distribution

    def values(self) -> t.Union[t.List[float], t.List[int]]:
        return [
            element.value
            for element in getattr(
                self.protobuf(),
                t.cast(str, self.protobuf().WhichOneof("distribution")),
            ).points
        ]

    def probabilities(self) -> t.List[float]:
        return [
            element.probability
            for element in getattr(
                self.protobuf(),
                t.cast(str, self.protobuf().WhichOneof("distribution")),
            ).points
        ]

    def names(self) -> t.Union[t.List[bool], t.List[str]]:
        distrib_type = t.cast(str, self.protobuf().WhichOneof("distribution"))
        if distrib_type in ["integer", "double"]:
            raise TypeError(
                f"{distrib_type} distribution has no attribute names"
            )
        return [
            element.name
            for element in getattr(self.protobuf(), distrib_type).points
        ]

    def min_value(self) -> t.Union[int, float]:
        distrib_type = t.cast(str, self.protobuf().WhichOneof("distribution"))
        if distrib_type in ["enum", "boolean"]:
            raise TypeError(f"{distrib_type} distribution has no min")
        # mypy thinks the return should be Any
        if distrib_type == "float":
            return t.cast(float, getattr(self.protobuf(), distrib_type).min)
        return t.cast(int, getattr(self.protobuf(), distrib_type).min)

    def max_value(self) -> t.Union[int, float]:
        distrib_type = t.cast(str, self.protobuf().WhichOneof("distribution"))
        if distrib_type in ["enum", "boolean"]:
            raise TypeError(f"{distrib_type} distribution has no max")
        # mypy thinks the return should be Any
        if distrib_type == "float":
            return t.cast(float, getattr(self.protobuf(), distrib_type).max)
        return t.cast(int, getattr(self.protobuf(), distrib_type).max)


# A few builders


# Distributions
def Integer_Distribution(
    min_value: t.Optional[int] = None,
    max_value: t.Optional[int] = None,
    probabilities: t.Optional[t.List[float]] = None,
    values: t.Optional[t.List[int]] = None,
) -> Distribution:
    if min_value is None:
        min_value = np.iinfo(np.int64).min
    if max_value is None:
        max_value = np.iinfo(np.int64).max
    if probabilities is None:
        probabilities = []
    if values is None:
        values = []
    return Distribution(
        sp.Distribution(
            integer=sp.Distribution.Integer(
                points=[
                    sp.Distribution.Integer.Point(
                        value=value, probability=probability
                    )
                    for value, probability in zip(values, probabilities)
                ],
                min=min_value,
                max=max_value,
            )
        )
    )


def Boolean_Distribution(
    probabilities: t.Optional[t.List[float]] = None,
    names: t.Optional[t.List[bool]] = None,
    values: t.Optional[t.List[int]] = None,
) -> Distribution:
    if probabilities is None:
        probabilities = []
    if names is None:
        names = []
    if values is None:
        values = []
    return Distribution(
        sp.Distribution(
            boolean=sp.Distribution.Boolean(
                points=[
                    sp.Distribution.Boolean.Point(
                        name=name, value=value, probability=probability
                    )
                    for name, value, probability in zip(
                        names, values, probabilities
                    )
                ]
            )
        )
    )


def Double_Distribution(
    min_value: t.Optional[float] = None,
    max_value: t.Optional[float] = None,
    probabilities: t.Optional[t.List[float]] = None,
    values: t.Optional[t.List[float]] = None,
) -> Distribution:
    if min_value is None:
        min_value = np.finfo(np.float64).min  # type:ignore
    if max_value is None:
        max_value = np.finfo(np.float64).max  # type:ignore
    if probabilities is None:
        probabilities = []
    if values is None:
        values = []
    return Distribution(
        sp.Distribution(
            double=sp.Distribution.Double(
                points=[
                    sp.Distribution.Double.Point(
                        value=value, probability=probability
                    )
                    for value, probability in zip(values, probabilities)
                ],
                min=min_value,  # type:ignore
                max=max_value,  # type: ignore
            )
        )
    )


def Enum_Distribution(
    probabilities: t.Optional[t.List[float]] = None,
    names: t.Optional[t.List[str]] = None,
    values: t.Optional[t.List[float]] = None,
) -> Distribution:
    if probabilities is None:
        probabilities = []
    if names is None:
        names = []
    if values is None:
        values = [i for i in range(len(names))]
    return Distribution(
        sp.Distribution(
            enum=sp.Distribution.Enum(
                points=[
                    sp.Distribution.Enum.Point(
                        name=name, value=value, probability=probability
                    )
                    for name, value, probability in zip(
                        names, values, probabilities
                    )
                ]
            )
        )
    )


# Statistics
def Null(size: int, multiplicity: float) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Null",
            null=sp.Statistics.Null(size=size, multiplicity=multiplicity),
        )
    )


def Unit(
    size: int,
    multiplicity: float,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Unit",
            unit=sp.Statistics.Unit(size=size, multiplicity=multiplicity),
            properties=properties,
        )
    )


def Id(size: int, multiplicity: float) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Id",
            id=sp.Statistics.Id(size=size, multiplicity=multiplicity),
        )
    )


def Boolean(
    size: int,
    multiplicity: float,
    probabilities: t.Optional[t.List[float]] = None,
    names: t.Optional[t.List[bool]] = None,
    values: t.Optional[t.List[int]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Boolean",
            boolean=sp.Statistics.Boolean(
                distribution=Boolean_Distribution(
                    probabilities=probabilities, names=names, values=values
                ).protobuf(),
                size=size,
                multiplicity=multiplicity,
            ),
        )
    )


def Integer(
    size: int,
    multiplicity: float,
    min_value: t.Optional[int] = None,
    max_value: t.Optional[int] = None,
    probabilities: t.Optional[t.List[float]] = None,
    values: t.Optional[t.List[int]] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Integer",
            integer=sp.Statistics.Integer(
                distribution=Integer_Distribution(
                    min_value=min_value,
                    max_value=max_value,
                    probabilities=probabilities,
                    values=values,
                ).protobuf(),
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


def Enum(
    size: int,
    multiplicity: float,
    probabilities: t.Optional[t.List[float]] = None,
    names: t.Optional[t.List[str]] = None,
    values: t.Optional[t.List[float]] = None,
    name: str = "Enum",
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name=name,
            enum=sp.Statistics.Enum(
                distribution=Enum_Distribution(
                    probabilities=probabilities, values=values, names=names
                ).protobuf(),
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


def Float(
    size: int,
    multiplicity: float,
    min_value: t.Optional[float] = None,
    max_value: t.Optional[float] = None,
    probabilities: t.Optional[t.List[float]] = None,
    values: t.Optional[t.List[float]] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Float",
            float=sp.Statistics.Float(
                distribution=Double_Distribution(
                    min_value=min_value,
                    max_value=max_value,
                    probabilities=probabilities,
                    values=values,
                ).protobuf(),
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


def Text(
    size: int,
    multiplicity: float,
    example: str = "",
    min_value: t.Optional[int] = None,
    max_value: t.Optional[int] = None,
    probabilities: t.Optional[t.List[float]] = None,
    values: t.Optional[t.List[int]] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Text",
            text=sp.Statistics.Text(
                distribution=Integer_Distribution(
                    min_value=min_value,
                    max_value=max_value,
                    probabilities=probabilities,
                    values=values,
                ).protobuf(),
                example=example,
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


def Bytes(size: int, multiplicity: float) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Bytes",
            bytes=sp.Statistics.Bytes(size=size, multiplicity=multiplicity),
        )
    )


def Struct(
    fields: t.Mapping[str, st.Statistics],
    size: int,
    multiplicity: float,
    name: t.Optional[str] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Struct" if name is None else name,
            struct=sp.Statistics.Struct(
                fields=[
                    sp.Statistics.Struct.Field(
                        name=name, statistics=statistics.protobuf()
                    )
                    for name, statistics in fields.items()
                ],
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


def Union(
    fields: t.Mapping[str, st.Statistics],
    size: int,
    multiplicity: float,
    name: str = "Union",
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name=name,
            union=sp.Statistics.Union(
                fields=[
                    sp.Statistics.Union.Field(
                        name=field_name, statistics=field_stat.protobuf()
                    )
                    for field_name, field_stat in fields.items()
                ],
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


def Optional(
    statistics: st.Statistics,
    size: int,
    multiplicity: float,
    name: str = "Optional",
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name=name,
            optional=sp.Statistics.Optional(
                statistics=statistics.protobuf(),
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


def List(
    statistics: st.Statistics,
    size: int,
    multiplicity: float,
    name: str = "List",
    min_value: t.Optional[int] = None,
    max_value: t.Optional[int] = None,
    probabilities: t.Optional[t.List[float]] = None,
    values: t.Optional[t.List[int]] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name=name,
            list=sp.Statistics.List(
                statistics=statistics.protobuf(),
                distribution=Integer_Distribution(
                    min_value=min_value,
                    max_value=max_value,
                    probabilities=probabilities,
                    values=values,
                ).protobuf(),
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


def Array(
    statistics: st.Statistics,
    size: int,
    multiplicity: float,
    name: str = "Array",
    min_values: t.Optional[t.List[float]] = None,
    max_values: t.Optional[t.List[float]] = None,
    probabilities: t.Optional[t.List[t.List[float]]] = None,
    values: t.Optional[t.List[t.List[float]]] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    if min_values is None:
        min_values = []
    if max_values is None:
        max_values = []
    if probabilities is None:
        probabilities = [[]]
    if values is None:
        values = [[]]
    distributions = [
        sp.Distribution(
            double=sp.Distribution.Double(
                points=[
                    sp.Distribution.Double.Point(
                        value=value, probability=probability
                    )
                    for value, probability in zip(
                        value_dimension, probability_dimension
                    )
                ],
                min=min_value,
                max=max_value,
            )
        )
        for value_dimension, probability_dimension, min_value, max_value in zip(  # noqa: E501
            values, probabilities, min_values, max_values
        )
    ]
    return Statistics(
        sp.Statistics(
            name=name,
            array=sp.Statistics.Array(
                statistics=statistics.protobuf(),
                distributions=distributions,
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


def Datetime(
    size: int,
    multiplicity: float,
    min_value: t.Optional[int] = None,
    max_value: t.Optional[int] = None,
    probabilities: t.Optional[t.List[float]] = None,
    values: t.Optional[t.List[int]] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Datetime",
            datetime=sp.Statistics.Datetime(
                distribution=Integer_Distribution(
                    min_value=min_value,
                    max_value=max_value,
                    probabilities=probabilities,
                    values=values,
                ).protobuf(),
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


def Date(
    size: int,
    multiplicity: float,
    min_value: t.Optional[int] = None,
    max_value: t.Optional[int] = None,
    probabilities: t.Optional[t.List[float]] = None,
    values: t.Optional[t.List[int]] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Date",
            date=sp.Statistics.Date(
                distribution=Integer_Distribution(
                    min_value=min_value,
                    max_value=max_value,
                    probabilities=probabilities,
                    values=values,
                ).protobuf(),
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


def Time(
    size: int,
    multiplicity: float,
    min_value: t.Optional[int] = None,
    max_value: t.Optional[int] = None,
    probabilities: t.Optional[t.List[float]] = None,
    values: t.Optional[t.List[int]] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Time",
            time=sp.Statistics.Time(
                distribution=Integer_Distribution(
                    min_value=min_value,
                    max_value=max_value,
                    probabilities=probabilities,
                    values=values,
                ).protobuf(),
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


def Duration(
    size: int,
    multiplicity: float,
    min_value: t.Optional[int] = None,
    max_value: t.Optional[int] = None,
    probabilities: t.Optional[t.List[float]] = None,
    values: t.Optional[t.List[int]] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Duration",
            duration=sp.Statistics.Duration(
                distribution=Integer_Distribution(
                    min_value=min_value,
                    max_value=max_value,
                    probabilities=probabilities,
                    values=values,
                ).protobuf(),
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


def Constrained(
    statistics: st.Statistics,
    size: int,
    multiplicity: float,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Statistics:
    return Statistics(
        sp.Statistics(
            name="Constrained",
            constrained=sp.Statistics.Constrained(
                statistics=statistics.protobuf(),
                size=size,
                multiplicity=multiplicity,
            ),
            properties=properties,
        )
    )


# A New Visitor base implementation
class StatisticsVisitor(st.StatisticsVisitor):
    """A base implementation for visitor class"""

    def default(self) -> None:
        raise NotImplementedError

    def Null(self, size: int, multiplicity: float) -> None:
        self.default()

    def Unit(self, size: int, multiplicity: float) -> None:
        self.default()

    def Boolean(
        self,
        size: int,
        multiplicity: float,
        probabilities: t.Optional[t.List[float]] = None,
        names: t.Optional[t.List[bool]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None:
        self.default()

    def Id(self, size: int, multiplicity: float) -> None:
        self.default()

    def Integer(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None:
        self.default()

    def Enum(
        self,
        size: int,
        multiplicity: float,
        probabilities: t.Optional[t.List[float]] = None,
        names: t.Optional[t.List[str]] = None,
        values: t.Optional[t.List[float]] = None,
        name: str = "Enum",
    ) -> None:
        self.default()

    def Float(
        self,
        size: int,
        multiplicity: float,
        min_value: float,
        max_value: float,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[float]] = None,
    ) -> None:
        self.default()

    def Text(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        example: str = "",
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None:
        self.default()

    def Bytes(self, size: int, multiplicity: float) -> None:
        self.default()

    def Struct(
        self,
        fields: t.Mapping[str, st.Statistics],
        size: int,
        multiplicity: float,
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None:
        self.default()

    def Union(
        self,
        fields: t.Mapping[str, st.Statistics],
        size: int,
        multiplicity: float,
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None:
        self.default()

    def Optional(
        self, statistics: st.Statistics, size: int, multiplicity: float
    ) -> None:
        self.default()

    def List(
        self,
        statistics: st.Statistics,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        name: str = "List",
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None:
        self.default()

    def Array(
        self,
        statistics: st.Statistics,
        size: int,
        multiplicity: float,
        min_values: t.Optional[t.List[float]] = None,
        max_values: t.Optional[t.List[float]] = None,
        name: str = "Array",
        probabilities: t.Optional[t.List[t.List[float]]] = None,
        values: t.Optional[t.List[t.List[float]]] = None,
    ) -> None:
        self.default()

    def Datetime(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None:
        self.default()

    def Time(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None:
        self.default()

    def Date(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None:
        self.default()

    def Duration(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None:
        self.default()

    def Constrained(
        self, statistics: st.Statistics, size: int, multiplicity: float
    ) -> None:
        self.default()


if t.TYPE_CHECKING:
    test_stat: st.Statistics = Statistics(sp.Statistics())
    test_dis: st.Distribution = Distribution(sp.Distribution())



"""Protocols describing common object behaviors."""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum, IntEnum
import datetime as dt
import logging
import typing as t
import warnings

import numpy as np
import pandas as pd
import pyarrow as pa

from sarus_data_spec.protobuf.typing import Protobuf, ProtobufWithUUID
import sarus_data_spec.manager.typing as manager_typing
import sarus_data_spec.protobuf as sp
import sarus_data_spec.storage.typing as storage_typing

logger = logging.getLogger(__name__)

try:
    import tensorflow as tf
except ModuleNotFoundError:
    logger.warning("tensorflow not found, tensorflow datasets not available")

try:
    import sklearn  # noqa: F401
except ModuleNotFoundError:
    logger.warning("sklearn not found, sklearn models not available")

try:
    from sarus_differential_privacy.query import (
        PrivateQuery as RealPrivateQuery,
    )

    PrivateQuery = RealPrivateQuery
except ImportError:
    PrivateQuery = t.Any  # type: ignore
    warnings.warn(
        "`sarus_differential_privacy` not installed. "
        "DP primitives not available."
    )

# this is really bad but in QB actually it is not defined!
Task = t.Any

try:
    import transformers  # noqa: F401
except ImportError:
    warnings.warn("`transformers` not installed. LLM not available.")


if t.TYPE_CHECKING:
    from sklearn import svm

    DataSpecValue = t.Union[pd.DataFrame, np.ndarray, svm.SVC]
else:
    DataSpecValue = t.Any


# List of types a Dataset can be converted to
DatasetCastable = t.Union[
    pd.DataFrame,
    pd.Series,
    t.Iterator[pa.RecordBatch],
    t.AsyncIterator[pa.RecordBatch],
    pd.core.groupby.DataFrameGroupBy,
    pd.core.groupby.SeriesGroupBy,
]


P = t.TypeVar("P", bound=Protobuf, covariant=True)


@t.runtime_checkable
class HasProtobuf(t.Protocol[P]):
    """An object backed by a protocol buffer message."""

    def protobuf(self) -> P:
        """Returns the underlying protobuf object."""
        ...

    def prototype(self) -> t.Type[P]:
        """Returns the type of protobuf."""
        ...

    def type_name(self) -> str:
        """Returns the name of the type."""
        ...

    def __getitem__(self, key: str) -> str:
        """Returns the property referred by key"""
        ...

    def properties(self) -> t.Mapping[str, str]:
        """Returns the properties"""
        ...


@t.runtime_checkable
class Value(t.Protocol):
    """An object with value semantics."""

    def __bytes__(self) -> bytes: ...

    def __repr__(self) -> str: ...

    def __str__(self) -> str: ...

    def __eq__(self, value: object) -> bool: ...

    def __hash__(self) -> int: ...


@t.runtime_checkable
class Frozen(t.Protocol):
    """An immutable object."""

    def _freeze(self) -> None:
        """Freeze the state of the object"""
        ...

    def _frozen(self) -> bool:
        """Check if the frozen object was left unchanged"""
        ...


PU = t.TypeVar("PU", bound=ProtobufWithUUID, covariant=True)


@t.runtime_checkable
class Referrable(HasProtobuf[PU], Frozen, t.Protocol[PU]):
    """Can be referred to by uuid."""

    def uuid(self) -> str:
        """Reference to use to refer to this object."""
        ...

    def referring(
        self, type_name: t.Optional[str] = None
    ) -> t.Collection[Referring[ProtobufWithUUID]]:
        """Referring objects pointing to this one."""
        ...

    def storage(self) -> storage_typing.Storage: ...

    def manager(self) -> manager_typing.Manager: ...


@t.runtime_checkable
class Referring(Referrable[PU], Frozen, t.Protocol[PU]):
    """Is referring to other Referrables"""

    _referred: t.MutableSet[str] = set()

    def referred(self) -> t.Collection[Referrable[ProtobufWithUUID]]:
        """Referred by this object."""
        ...

    def referred_uuid(self) -> t.Collection[str]:
        """Uuid of object referred by this object"""
        ...


FM = t.TypeVar("FM", bound=Protobuf)


@t.runtime_checkable
class Factory(t.Protocol):
    """Can produce objects from protobuf messages"""

    def register(self, name: str, type: t.Type[HasProtobuf[Protobuf]]) -> None:
        """Registers a class"""
        ...

    def create(self, message: FM, store: bool) -> HasProtobuf[FM]:
        """Returns a wrapped protobuf"""
        ...


class VariantConstraint(Referrable[sp.VariantConstraint], t.Protocol):
    def constraint_kind(self) -> ConstraintKind: ...

    def required_context(self) -> t.List[str]: ...

    def privacy_limit(self) -> t.Optional[PrivacyLimit]: ...

    def accept(self, visitor: TransformVisitor) -> None: ...


# Type alias
DS = t.TypeVar("DS", bound=t.Union[sp.Scalar, sp.Dataset])


class Attribute(Referring[sp.Attribute], t.Protocol):
    def prototype(self) -> t.Type[sp.Attribute]: ...

    def name(self) -> str: ...


@t.runtime_checkable
class DataSpec(Referring[DS], t.Protocol):
    def parents(
        self,
    ) -> t.Tuple[
        t.List[t.Union[DataSpec[DS], Transform]],
        t.Dict[str, t.Union[DataSpec[DS], Transform]],
    ]: ...

    def parents_list(
        self,
    ) -> t.List[DataSpec[DS]]: ...

    def variant(
        self,
        kind: ConstraintKind,
        public_context: t.Collection[str] = (),
        privacy_limit: t.Optional[PrivacyLimit] = None,
        salt: t.Optional[int] = None,
    ) -> t.Optional[DataSpec[DS]]: ...

    def variants(self) -> t.Collection[DataSpec]: ...

    def private_queries(self) -> t.List[PrivateQuery]:
        """Return the list of PrivateQueries used in a Dataspec's transform.

        It represents the privacy loss associated with the current computation.

        It can be used by Sarus when a user (Access object) reads a DP dataspec
        to update its accountant. Note that Private Query objects are generated
        with a random uuid so that even if they are submitted multiple times to
        an account, they are only accounted once (ask @cgastaud for more on
        accounting)."""
        ...

    def name(self) -> str: ...

    def doc(self) -> str: ...

    def is_pup(self) -> bool: ...

    def pup_token(self) -> t.Optional[str]:
        """Returns a PUP token if the dataset is PUP and None otherwise.

        The PUP token is stored in the properties of the VariantConstraint. It
        is a hash initialized with a value when the Dataset is protected.

        If a transform does not preserve the PEID then the token is set to None
        If a transform preserves the PEID assignment but changes the rows (e.g.
        sample, shuffle, filter,...) then the token's value is changed If a
        transform does not change the rows (e.g. selecting a column, adding a
        scalar,...) then the token is passed without change

        A Dataspec is PUP if its PUP token is not None. Two PUP Dataspecs are
        aligned (i.e. they have the same number of rows and all their rows have
        the same PEID) if their tokens are equal.
        """
        ...

    def rewritten_pup_token(self) -> t.Optional[str]:
        """Returns the PUP token for the DP variant of this dataspec
        during DP rewriting.

        Currently, the implementation assumes a single DP/PUP variant per
        dataspec, resulting in one possible value for
        "rewritten_pup_token."
        Future changes could introduce multiple variants, necessitating
        the implementation of priority rules.
        """
        ...

    def is_public(self) -> bool: ...

    def is_synthetic(self) -> bool:
        """Is the dataspec synthetic."""
        ...

    def is_dp(self) -> bool:
        """Is the dataspec the result of a DP transform."""
        ...

    def is_dp_writable(self) -> bool:
        """Check if it exists a variant of this dataspec
        that utilizes the dp equivalent, and this variant is is_dp."""
        ...

    def is_pup_writable(self) -> bool: ...

    def is_publishable(self) -> bool: ...

    def is_published(self) -> bool:
        """Check if the dataspec is the result of a DP transform or another
        published dataspec.
        There is at least one parent that is DP.
        Such a dataspec has at least one private query in its computation
        graph."""
        ...

    def is_transformed(self) -> bool:
        """Is the dataspec transformed."""
        ...

    def is_source(self) -> bool:
        """Is the dataspec a source dataspec."""
        ...

    def is_big_data_computable(self) -> bool:
        """Check if the dataspec can be computed if it is big data."""
        ...

    def is_remote(self) -> bool:
        """Is the dataspec a remotely defined dataset."""
        ...

    def sources(self, type_name: t.Optional[str]) -> t.Set[DataSpec]: ...

    def transform(self) -> Transform: ...

    def status(
        self, task_names: t.Optional[t.List[str]]
    ) -> t.Optional[Status]: ...

    def accept(self, visitor: Visitor) -> None: ...

    def attribute(self, name: str) -> t.Optional[Attribute]:
        """Return the attribute with the given name or None if not found."""
        ...

    def attributes(self, name: str) -> t.List[Attribute]:
        """Return all the attributes with the given name."""
        ...


class Dataset(DataSpec[sp.Dataset], t.Protocol):
    def prototype(self) -> t.Type[sp.Dataset]: ...

    def is_synthetic(self) -> bool: ...

    def has_admin_columns(self) -> bool: ...

    def is_protected(self) -> bool: ...

    def is_file(self) -> bool: ...

    def schema(self) -> Schema: ...

    async def async_schema(self) -> Schema: ...

    def size(self) -> t.Optional[Size]: ...

    def multiplicity(self) -> t.Optional[Multiplicity]: ...

    def bounds(self) -> t.Optional[Bounds]: ...

    def marginals(self) -> t.Optional[Marginals]: ...

    def to_arrow(
        self, batch_size: int = 10000
    ) -> t.Iterator[pa.RecordBatch]: ...

    async def async_to_arrow(
        self, batch_size: int = 10000
    ) -> t.AsyncIterator[pa.RecordBatch]: ...

    def to_sql(self) -> None: ...

    def push_sql(self) -> None: ...

    def spec(self) -> str: ...

    def __iter__(self) -> t.Iterator[pa.RecordBatch]: ...

    def to_pandas(self) -> pd.DataFrame: ...

    async def async_to_pandas(self) -> pd.DataFrame: ...

    def to_tensorflow(self) -> tf.data.Dataset: ...

    async def async_to_tensorflow(self) -> tf.data.Dataset: ...

    async def async_to(
        self, kind: t.Type, drop_admin: bool = True
    ) -> DatasetCastable:
        """Casts a Dataset to a Python type passed as argument."""
        ...

    def to(self, kind: t.Type, drop_admin: bool = True) -> DatasetCastable: ...

    def dot(self) -> str:
        """return a graphviz representation of the dataset"""
        ...

    def sql(
        self,
        query: t.Union[str, t.Dict[str, t.Any]],
        dialect: t.Optional[SQLDialect] = None,
        batch_size: int = 10000,
    ) -> t.Iterator[pa.RecordBatch]:
        """Executes the sql method on the dataset"""
        ...

    def foreign_keys(self) -> t.Dict[Path, Path]:
        """returns foreign keys of the dataset"""
        ...

    def primary_keys(self) -> t.List[Path]:
        """Returns a list of the paths to all primary keys"""
        ...

    def links(self) -> Links:
        """Returns the foreign keys
        distributions of the dataset computed
        with dp"""
        ...


class Scalar(DataSpec[sp.Scalar], t.Protocol):
    def prototype(self) -> t.Type[sp.Scalar]:
        """Return the type of the underlying protobuf."""
        ...

    def is_pretrained_model(self) -> bool: ...

    def is_privacy_params(self) -> bool:
        """Is the scalar privacy parameters."""

    def is_random_seed(self) -> bool:
        """Is the scalar a random seed."""

    def is_synthetic_model(self) -> bool:
        """is the scalar a synthetic model"""

    def value(self) -> DataSpecValue: ...

    async def async_value(self) -> DataSpecValue: ...

    def spec(self) -> str: ...


class Visitor(t.Protocol):
    """A visitor class for Dataset"""

    def all(self, visited: DataSpec) -> None: ...

    def transformed(
        self,
        visited: DataSpec,
        transform: Transform,
        *arguments: DataSpec,
        **named_arguments: DataSpec,
    ) -> None: ...

    def other(self, visited: DataSpec) -> None: ...


class Bounds(Referring[sp.Bounds], t.Protocol):
    """A python abstract class to describe bounds"""

    def prototype(self) -> t.Type[sp.Bounds]:
        """Return the type of the underlying protobuf."""
        ...

    def dataset(self) -> Dataset: ...

    def statistics(self) -> Statistics: ...


class Marginals(Referring[sp.Marginals], t.Protocol):
    """A python abstract class to describe marginals"""

    def prototype(self) -> t.Type[sp.Marginals]:
        """Return the type of the underlying protobuf."""
        ...

    def dataset(self) -> Dataset: ...

    def statistics(self) -> Statistics: ...


class Size(Referring[sp.Size], t.Protocol):
    """A python abstract class to describe size"""

    def prototype(self) -> t.Type[sp.Size]:
        """Return the type of the underlying protobuf."""
        ...

    def dataset(self) -> Dataset: ...

    def statistics(self) -> Statistics: ...


class Multiplicity(Referring[sp.Multiplicity], t.Protocol):
    """A python abstract class to describe size"""

    def prototype(self) -> t.Type[sp.Multiplicity]:
        """Return the type of the underlying protobuf."""
        ...

    def dataset(self) -> Dataset: ...

    def statistics(self) -> Statistics: ...


class Schema(Referring[sp.Schema], t.Protocol):
    """A python abstract class to describe schemas"""

    def prototype(self) -> t.Type[sp.Schema]:
        """Return the type of the underlying protobuf."""
        ...

    def name(self) -> str: ...

    def dataset(self) -> Dataset: ...

    def to_arrow(self) -> pa.Schema: ...

    def type(self) -> Type: ...

    def has_admin_columns(self) -> bool: ...

    def is_privacy_unit_tracking(self) -> bool: ...

    def tables(self) -> t.List[Path]: ...

    def privacy_unit_tracking_paths(self) -> Path: ...

    def data_type(self) -> Type: ...

    def private_tables(self) -> t.List[Path]: ...

    def public_tables(self) -> t.List[Path]: ...


class Status(Referring[sp.Status], t.Protocol):
    """A python abstract class to describe status"""

    def prototype(self) -> t.Type[sp.Status]:
        """Return the type of the underlying protobuf."""
        ...

    def dataspec(self) -> DataSpec: ...

    def datetime(self) -> dt.datetime: ...

    def update(
        self,
        task_stages: t.Optional[t.Mapping[str, Stage]],
        properties: t.Optional[t.Mapping[str, str]],
    ) -> t.Tuple[Status, bool]: ...

    def task(self, task: str) -> t.Optional[Stage]: ...

    def pending(self) -> bool: ...

    def processing(self) -> bool: ...

    def ready(self) -> bool: ...

    def error(self) -> bool: ...

    def owner(
        self,
    ) -> manager_typing.Manager:  # TODO: Maybe find a better name, but this was shadowing the actual manager of this object.  # noqa: E501
        ...

    def clear_task(self, task: str) -> t.Tuple[Status, bool]:
        """Creates a new status removing the task specified.
        If the task does not exist, nothing is created"""


class Stage(HasProtobuf[sp.Status.Stage], t.Protocol):
    def accept(self, visitor: StageVisitor) -> None: ...

    def stage(self) -> str: ...

    def ready(self) -> bool: ...

    def processing(self) -> bool: ...

    def pending(self) -> bool: ...

    def error(self) -> bool: ...


class StageVisitor(t.Protocol):
    """A visitor class for Status/Stage"""

    def pending(self) -> None: ...

    def processing(self) -> None: ...

    def ready(self) -> None: ...

    def error(self) -> None: ...


@t.runtime_checkable
class Transform(Referrable[sp.Transform], t.Protocol):
    """A python abstract class to describe transforms"""

    def prototype(self) -> t.Type[sp.Transform]:
        """Return the type of the underlying protobuf."""
        ...

    def name(self) -> str: ...

    def doc(self) -> str: ...

    def spec(self) -> str: ...

    def is_composed(self) -> bool:
        """Is the transform composed."""
        ...

    def is_variable(self) -> bool:
        """Is the transform a variable."""
        ...

    def is_external(self) -> bool:
        """Is the transform an external operation."""
        ...

    def infer_output_type(
        self,
        *arguments: t.Union[DataSpec, Transform],
        **named_arguments: t.Union[DataSpec, Transform],
    ) -> t.Tuple[str, t.Callable[[DataSpec], None]]:
        """Guess if the external transform output is a Dataset or a Scalar.

        Registers schema if it is a Dataset and returns the value type.
        """
        ...

    def transforms(self) -> t.Set[Transform]:
        """return all transforms (and avoid infinite recursions/loops)"""
        ...

    def variables(self) -> t.Set[Transform]:
        """Return all the variables from a composed transform"""
        ...

    def compose(
        self,
        *compose_arguments: Transform,
        **compose_named_arguments: Transform,
    ) -> Transform: ...

    def apply(
        self,
        *apply_arguments: DataSpec,
        **apply_named_arguments: DataSpec,
    ) -> DataSpec: ...

    def abstract(
        self,
        *arguments: str,
        **named_arguments: str,
    ) -> Transform: ...

    def __call__(
        self,
        *arguments: t.Union[Transform, DataSpec, int, str],
        **named_arguments: t.Union[Transform, DataSpec, int, str],
    ) -> t.Union[Transform, DataSpec]:
        """Applies the transform to another element"""
        ...

    def __mul__(self, argument: Transform) -> Transform: ...

    def accept(self, visitor: TransformVisitor) -> None: ...

    def transform_to_apply(self) -> Transform:
        """Return the transform of a composed transform."""

    def composed_parents(
        self,
    ) -> t.Tuple[t.List[Transform], t.Dict[str, Transform]]:
        """Return the parents of a composed transform."""

    def composed_callable(self) -> t.Callable[..., t.Any]:
        """Return the composed transform's equivalent callable.

        The function takes an undefined number of named arguments.
        """


class TransformVisitor(t.Protocol):
    """A visitor class for Transform"""

    def all(self, visited: Transform) -> None: ...

    def composed(
        self,
        visited: Transform,
        transform: Transform,
        *arguments: Transform,
        **named_arguments: Transform,
    ) -> None: ...

    def variable(
        self,
        visited: Transform,
        name: str,
        position: int,
    ) -> None: ...

    def other(self, visited: Transform) -> None: ...


class Path(HasProtobuf[sp.Path], Frozen, Value, t.Protocol):
    """
    A python class to describe Paths
    A Path is essentially a tree of paths
    Each Path in the tree is a sequence of Labels
    """

    def prototype(self) -> t.Type[sp.Path]:
        """Return the type of the underlying protobuf."""
        ...

    def label(self) -> str: ...

    def sub_paths(self) -> t.List[Path]:
        """Select within self the suffix paths that have
        select_path as their prefixes
        """
        ...

    def is_empty(self) -> bool:
        """Tells if the path is empty (no label and no subpath)"""
        ...

    def has_sub_paths(self) -> bool:
        """Tells if the path has sub-paths"""
        ...

    def select(self, select_path: Path) -> t.List[Path]: ...

    def to_strings_list(self) -> t.List[t.List[str]]: ...

    def to_dict(self) -> t.Dict[str, str]: ...


class Type(HasProtobuf[sp.Type], Frozen, Value, t.Protocol):
    def prototype(self) -> t.Type[sp.Type]:
        """Return the type of the underlying protobuf."""
        ...

    def name(self) -> str:
        """Returns the name of the underlying protobuf."""
        ...

    def data_type(self) -> Type:
        """Returns the first type level containing the data,
        hence skips the protected_entity struct if there is one"""

    def has_admin_columns(self) -> bool:
        """Return True if the Type has administrative columns."""

    def has_privacy_unit_tracking(self) -> bool:
        """Return True if the Type has protection information."""

    def structs(self: Type) -> t.Optional[t.List[Path]]:
        """Returns the path to the first level structs encountered in the
        type.
        For example, Union[Struct1,Union[Struct2[Struct3]] will return only a
        path that brings to Struct1 and Struct2.
        """
        ...

    def get(self, item: Path) -> Type:
        """Return a subtype of the considered type defined by the path."""
        ...

    def leaves(self) -> t.List[Type]:
        """Returns the leaves contained in the type tree structure"""
        ...

    def children(self) -> t.Dict[str, Type]:
        """Returns the children contained in the type tree structure"""
        ...

    # A Visitor acceptor
    def accept(self, visitor: TypeVisitor) -> None: ...

    # A new Visitor acceptor
    def accept_(self, visitor: TypeVisitor_[VisitedType]) -> VisitedType: ...

    def sub_types(self: Type, item: Path) -> t.List[Type]:
        """Returns the terminal nodes contained in the path"""
        ...

    def default(self: Type) -> pa.Array:
        """Returns an example of arrow array matching the type.
        For an optional type, it sets the default missing value.
        """

    def numpy_default(self: Type) -> np.ndarray:
        """Returns an example of numpy array matching the type.
        For an optional type, it sets the default missing value
        """

    def tensorflow_default(self, is_optional: bool = False) -> t.Any:
        """This methods returns a dictionary with tensors as leaves
        that match the type. For an optional type, we consider
        the case where the field is missing, and set the default value
        for each missing type.
        """

    def example(self: Type) -> pa.Array:
        """Returns an example of arrow array matching the type.
        For an optional type, it returns a non missing
        value of the type.
        """

    def numpy_example(self: Type) -> np.ndarray:
        """Returns an example of numpy array matching the type.
        For an optional type, it returns a non
        missing value of the type.
        """

    def tensorflow_example(self: Type) -> t.Any:
        """Returns an example of a dictionary with tensors as leaves
        that match the type..
        For an optional type, it returns a non missing value of the type.
        """

    def path_leaves(self) -> t.Sequence[Path]:
        """Returns the list of each path to a leaf in the type. If the type
        is a leaf, it returns an empty list"""


class IdBase(IntEnum):
    INT64 = 0
    INT32 = 1
    INT16 = 2
    INT8 = 3
    STRING = 4
    BYTES = 5


class DatetimeBase(IntEnum):
    INT64_NS = 0
    INT64_MS = 1
    STRING = 2


class DateBase(IntEnum):
    INT32 = 0
    STRING = 1


class TimeBase(IntEnum):
    INT64_NS = 0
    INT32_MS = 1
    STRING = 2
    INT64_US = 3


class IntegerBase(IntEnum):
    INT64 = 0
    INT32 = 1
    INT16 = 2
    INT8 = 3
    UINT64 = 4
    UINT32 = 5
    UINT16 = 6
    UINT8 = 7


class FloatBase(IntEnum):
    FLOAT64 = 0
    FLOAT32 = 1
    FLOAT16 = 2


class ConstraintKind(Enum):
    SYNTHETIC = sp.ConstraintKind.SYNTHETIC
    PUP = sp.ConstraintKind.PUP
    DP = sp.ConstraintKind.DP
    PUBLIC = sp.ConstraintKind.PUBLIC
    MOCK = sp.ConstraintKind.MOCK
    PUP_FOR_REWRITING = sp.ConstraintKind.PUP_FOR_REWRITING


class SQLDialect(Enum):
    """SQL Dialects"""

    POSTGRES = 1
    SQL_SERVER = 2
    MY_SQL = 3
    SQLLITE = 4
    ORACLE = 5
    BIG_QUERY = 6
    REDSHIFT = 7
    HIVE = 8
    DATABRICKS = 9
    MONGODB = 10


class InferredDistributionName(Enum):
    UNIFORM = "Uniform"
    NORMAL = "Normal"
    EXPONENTIAL = "Exponential"
    GAMMA = "Gamma"
    BETA = "Beta"
    PARETO = "Pareto"


class TypeVisitor(t.Protocol):
    """A visitor class for Type"""

    @abstractmethod
    def Null(
        self, properties: t.Optional[t.Mapping[str, str]] = None
    ) -> None: ...

    @abstractmethod
    def Unit(
        self, properties: t.Optional[t.Mapping[str, str]] = None
    ) -> None: ...

    @abstractmethod
    def Boolean(
        self, properties: t.Optional[t.Mapping[str, str]] = None
    ) -> None: ...

    @abstractmethod
    def Id(
        self,
        unique: bool,
        reference: t.Optional[Path] = None,
        base: t.Optional[IdBase] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Integer(
        self,
        min: int,
        max: int,
        base: IntegerBase,
        possible_values: t.Iterable[int],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Enum(
        self,
        name: str,
        name_values: t.Sequence[t.Tuple[str, int]],
        ordered: bool,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Float(
        self,
        min: float,
        max: float,
        base: FloatBase,
        possible_values: t.Iterable[float],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Text(
        self,
        encoding: str,
        possible_values: t.Iterable[str],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Bytes(
        self, properties: t.Optional[t.Mapping[str, str]] = None
    ) -> None: ...

    @abstractmethod
    def Struct(
        self,
        fields: t.Mapping[str, Type],
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Union(
        self,
        fields: t.Mapping[str, Type],
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Optional(
        self,
        type: Type,
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def List(
        self,
        type: Type,
        max_size: int,
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Array(
        self,
        type: Type,
        shape: t.Tuple[int, ...],
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Datetime(
        self,
        format: str,
        min: str,
        max: str,
        base: DatetimeBase,
        possible_values: t.Iterable[str],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Time(
        self,
        format: str,
        min: str,
        max: str,
        base: TimeBase,
        possible_values: t.Iterable[str],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Date(
        self,
        format: str,
        min: str,
        max: str,
        base: DateBase,
        possible_values: t.Iterable[str],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Duration(
        self,
        unit: str,
        min: int,
        max: int,
        possible_values: t.Iterable[int],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Constrained(
        self,
        type: Type,
        constraint: Predicate,
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Hypothesis(
        self,
        *types: t.Tuple[Type, float],
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...


VisitedType = t.TypeVar("VisitedType")


class TypeVisitor_(t.Protocol, t.Generic[VisitedType]):
    """A visitor class for Type"""

    def default(self) -> VisitedType:
        """This is the minimal implementation to provide"""
        ...

    def Null(
        self, properties: t.Optional[t.Mapping[str, str]] = None
    ) -> VisitedType: ...

    def Unit(
        self, properties: t.Optional[t.Mapping[str, str]] = None
    ) -> VisitedType: ...

    def Boolean(
        self, properties: t.Optional[t.Mapping[str, str]] = None
    ) -> VisitedType: ...

    def Id(
        self,
        unique: bool,
        reference: t.Optional[Path] = None,
        base: t.Optional[IdBase] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Integer(
        self,
        min: int,
        max: int,
        base: IntegerBase,
        possible_values: t.Iterable[int],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Enum(
        self,
        name: str,
        name_values: t.Sequence[t.Tuple[str, int]],
        ordered: bool,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Float(
        self,
        min: float,
        max: float,
        base: FloatBase,
        possible_values: t.Iterable[float],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Text(
        self,
        encoding: str,
        possible_values: t.Iterable[str],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Bytes(
        self, properties: t.Optional[t.Mapping[str, str]] = None
    ) -> VisitedType: ...

    def Struct(
        self,
        fields: t.Mapping[str, VisitedType],
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Union(
        self,
        fields: t.Mapping[str, VisitedType],
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Optional(
        self,
        type_: VisitedType,
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def List(
        self,
        type_: VisitedType,
        max_size: int,
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Array(
        self,
        type_: VisitedType,
        shape: t.Tuple[int, ...],
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Datetime(
        self,
        format: str,
        min: str,
        max: str,
        base: DatetimeBase,
        possible_values: t.Iterable[str],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Time(
        self,
        format: str,
        min: str,
        max: str,
        base: TimeBase,
        possible_values: t.Iterable[str],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Date(
        self,
        format: str,
        min: str,
        max: str,
        base: DateBase,
        possible_values: t.Iterable[str],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Duration(
        self,
        unit: str,
        min: int,
        max: int,
        possible_values: t.Iterable[int],
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Constrained(
        self,
        type_: VisitedType,
        constraint: Predicate,
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...

    def Hypothesis(
        self,
        *types: t.Tuple[VisitedType, float],
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> VisitedType: ...


class Predicate(HasProtobuf[sp.Predicate], Frozen, Value, t.Protocol):
    """A python class to describe types"""

    def prototype(self) -> t.Type[sp.Predicate]:
        """Return the type of the underlying protobuf."""

    # A bunch of operators
    def __or__(self, predicate: Predicate) -> Predicate:
        """Union operator"""

    def __and__(self, predicate: Predicate) -> Predicate:
        """Inter operator"""

    def __invert__(self) -> Predicate:
        """Complement"""


class Statistics(HasProtobuf[sp.Statistics], Frozen, Value, t.Protocol):
    """A python class to describe statistics"""

    def prototype(self) -> t.Type[sp.Statistics]:
        """Return the type of the underlying protobuf."""
        ...

    def name(self) -> str: ...

    def distribution(self) -> Distribution: ...

    def size(self) -> int: ...

    def multiplicity(self) -> float: ...

    def accept(self, visitor: StatisticsVisitor) -> None: ...

    def nodes_statistics(self, path: Path) -> t.List[Statistics]:
        """Returns the List of each statistics corresponding at the leaves
        of path"""
        ...

    def children(self) -> t.Dict[str, Statistics]:
        """Returns the children contained in the type tree structure"""
        ...


class Distribution(HasProtobuf[sp.Distribution], Frozen, Value, t.Protocol):
    """A python class to describe distributions"""

    def prototype(self) -> t.Type[sp.Distribution]:
        """Return the type of the underlying protobuf."""
        ...

    def values(self) -> t.Union[t.List[float], t.List[int]]: ...

    def probabilities(self) -> t.List[float]: ...

    def names(self) -> t.Union[t.List[bool], t.List[str]]: ...

    def min_value(self) -> t.Union[int, float]: ...

    def max_value(self) -> t.Union[int, float]: ...


class StatisticsVisitor(t.Protocol):
    """A visitor class for Statistics"""

    @abstractmethod
    def Null(self, size: int, multiplicity: float) -> None: ...

    @abstractmethod
    def Unit(self, size: int, multiplicity: float) -> None: ...

    def Boolean(
        self,
        size: int,
        multiplicity: float,
        probabilities: t.Optional[t.List[float]] = None,
        names: t.Optional[t.List[bool]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None: ...

    @abstractmethod
    def Id(self, size: int, multiplicity: float) -> None: ...

    @abstractmethod
    def Integer(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None: ...

    @abstractmethod
    def Enum(
        self,
        size: int,
        multiplicity: float,
        probabilities: t.Optional[t.List[float]] = None,
        names: t.Optional[t.List[str]] = None,
        values: t.Optional[t.List[float]] = None,
        name: str = "Enum",
    ) -> None: ...

    @abstractmethod
    def Float(
        self,
        size: int,
        multiplicity: float,
        min_value: float,
        max_value: float,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[float]] = None,
    ) -> None: ...

    @abstractmethod
    def Text(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        example: str = "",
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None: ...

    @abstractmethod
    def Bytes(self, size: int, multiplicity: float) -> None: ...

    @abstractmethod
    def Struct(
        self,
        fields: t.Mapping[str, Statistics],
        size: int,
        multiplicity: float,
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Union(
        self,
        fields: t.Mapping[str, Statistics],
        size: int,
        multiplicity: float,
        name: t.Optional[str] = None,
        properties: t.Optional[t.Mapping[str, str]] = None,
    ) -> None: ...

    @abstractmethod
    def Optional(
        self, statistics: Statistics, size: int, multiplicity: float
    ) -> None: ...

    @abstractmethod
    def List(
        self,
        statistics: Statistics,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        name: str = "List",
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None: ...

    @abstractmethod
    def Array(
        self,
        statistics: Statistics,
        size: int,
        multiplicity: float,
        min_values: t.Optional[t.List[float]] = None,
        max_values: t.Optional[t.List[float]] = None,
        name: str = "Array",
        probabilities: t.Optional[t.List[t.List[float]]] = None,
        values: t.Optional[t.List[t.List[float]]] = None,
    ) -> None: ...

    @abstractmethod
    def Datetime(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None: ...

    @abstractmethod
    def Date(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None: ...

    @abstractmethod
    def Time(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None: ...

    @abstractmethod
    def Duration(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None: ...

    @abstractmethod
    def Constrained(
        self, statistics: Statistics, size: int, multiplicity: float
    ) -> None: ...


class InferredDistribution(t.Protocol):
    """A python class to to infer user input distribution

    Attributes:
        nparams: number of parameters
    """

    nparams: int

    def estimate_params(self, x: np.ndarray) -> None:
        """estimate distribution parameters (non-DP) from data column"""
        ...

    def log_likelihood(self, x: np.ndarray) -> float:
        """compute log-likelihood of the distribution on data column"""
        ...

    def preprocess(self, x: np.ndarray) -> np.ndarray:
        """Shift/scale data to be able to estimate distribution parameters"""
        ...

    def params(self) -> t.Mapping[str, float]:
        """return distribution parameters"""
        ...


class InferredAlphabet(t.Protocol):
    """A python class to to infer user input charset

    Attributes:
        charset (t.List[int]): list with int representation of unique chars
        complexity (int): charset intervals used to generate the alphabet
            e.g. if alphabet (ascii) is [1,2,3, ..., 126] has complexity=1
            if alphabet is [1,2,3] U [10] = [1,2,3,10] has complexity=2
    """

    charset: t.List[int]
    complexity: int


T = t.TypeVar("T", covariant=True)


class Links(Referring[sp.Links], t.Protocol):
    """A python abstract class to describe all
    the links statistics in a dataset"""

    def prototype(self) -> t.Type[sp.Links]:
        """Return the type of the underlying protobuf."""
        ...

    def dataset(self) -> Dataset: ...

    def links_statistics(self) -> t.List[LinkStatistics]: ...


class LinkStatistics(HasProtobuf[sp.Links.LinkStat], t.Protocol):
    """A python class to describe the statistics of a link
    for a foreign key"""

    def prototype(self) -> t.Type[sp.Links.LinkStat]:
        """Return the type of the underlying protobuf."""
        ...

    def pointing(self) -> Path:
        """returns the path of the foreign key column"""
        ...

    def pointed(self) -> Path:
        """returns the path of the column pointed by a foreign_key"""
        ...

    def distribution(self) -> Distribution:
        """Returns the distribution of counts
        for the given pointing/pointed"""
        ...


class PrivacyLimit(t.Protocol):
    """An abstract Privacy Limit class."""

    def delta_epsilon_dict(self) -> t.Dict[float, float]:
        """Returns the limit as a dictionnary {delta: epsilon}"""
        pass



from __future__ import annotations

import shutil
import typing as t
import warnings
import logging
import pyarrow as pa
from sarus_data_spec.arrow.type import to_arrow
from sarus_data_spec.bounds import bounds as bounds_builder
from sarus_data_spec.constants import (
    DATASET_SLUGNAME,
    SYNTHETIC_MODEL,
    SYNTHETIC_TASK,
    DATA,
)
from pydantic import RootModel

import sarus_data_spec.type as sdt
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.marginals import marginals as marginals_builder
from sarus_data_spec.multiplicity import multiplicity as multiplicity_builder
from sarus_data_spec.size import size as size_builder
import sarus_data_spec.typing as st

try:
    from sarus_synthetic_data.generator import SyntheticDatasetGenerator
    from sarus_synthetic_data.configs.global_config import (
        SyntheticConfig,
        TrigramsColumn,
    )
except ModuleNotFoundError:
    warnings.warn("Synthetic Data not available")

from .config_updates import update_config

try:
    from sarus_data_spec.manager.ops.source.query_builder import (
        synthetic_parameters,
    )
except ModuleNotFoundError:
    warnings.warn(
        "synthetic_parameters not found, "
        "synthetic data operations not available "
    )
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.schema import schema
from sarus_data_spec.path import straight_path
from sarus_data_spec.manager.ops.processor.standard.standard_op import (
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
    StandardScalarImplementation,
    StandardScalarStaticChecker,
)
from sarus_data_spec.manager.ops.processor.standard.select_table import (
    select_table,
)
from sarus_data_spec.arrow.array import convert_record_batch
import os
import pyarrow.parquet as pq

logger = logging.getLogger(__name__)


def convert_array_to_table(
    schema_type: st.Type, arrow_data: pa.array
) -> pa.Array:
    """Given a PyArrow array, returns a correctly-defined Table."""

    class ArrayToTable(sdt.TypeVisitor):
        """Handles both configuration: a dataset as a Struct or as an Union."""

        result = None

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            names = list(fields.keys())
            self.result = pa.Table.from_arrays(
                arrays=arrow_data.flatten(), names=names
            )

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            names = list(fields.keys())
            arrs = arrow_data.flatten()
            schema = pa.schema(
                [
                    pa.field(name, type=arr.type)
                    for arr, name in zip(arrs[:-1], names)
                ]
            )
            schema = schema.append(
                pa.field(
                    "field_selected", type=pa.large_string(), nullable=False
                )
            )
            self.result = pa.Table.from_arrays(
                arrays=arrow_data.flatten(), schema=schema
            )

    visitor = ArrayToTable()
    schema_type.accept(visitor)
    return visitor.result


async def async_iter_arrow(
    iterator: t.Iterator[pa.RecordBatch],
) -> t.AsyncIterator[pa.RecordBatch]:
    """Async generator from the synthetic data iterator."""
    for batch in iterator:
        yield batch
    return


class SyntheticStaticChecker(StandardDatasetStaticChecker):
    def pup_token(self, public_context: t.Collection[str]) -> t.Optional[str]:
        # TODO add pup token when the synthetic data is actually protected
        return None

    def rewritten_pup_token(
        self, public_context: t.Collection[str]
    ) -> t.Optional[str]:
        # TODO add pup token when the synthetic data is actually protected
        return None

    async def schema(self) -> st.Schema:
        parent_schema = await self.parent_schema()
        return schema(
            self.dataset,
            schema_type=parent_schema.data_type(),
            properties=parent_schema.properties(),
            name=self.dataset.properties().get(DATASET_SLUGNAME, None),
        )


class Synthetic(StandardDatasetImplementation):
    """Create a Synthetic op class for is_pup."""

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        dataset = self.dataset
        parents, parents_dict = dataset.parents()

        # Forcing the marginals to be computed first
        parent = t.cast(Dataset, parents[0])
        marginals = await parent.manager().async_marginals(parent)
        schema = await parent.manager().async_schema(parent)
        parent_stats = marginals.statistics()

        # Budget
        budget_param = parents_dict["sd_budget"]
        budget = t.cast(
            t.Tuple[float, float],
            await dataset.manager().async_value(t.cast(Scalar, budget_param)),
        )

        # Model
        model_properties = t.cast(Scalar, parents_dict["synthetic_model"])
        synthetic_config = t.cast(
            SyntheticConfig,
            await dataset.manager().async_value(model_properties),
        )

        # Update SyntheticConfig with Calibrated Noise
        config_with_noise = await synthetic_parameters(
            dataset,
            sd_budget=budget,
            task=SYNTHETIC_TASK,
            synthetic_config=synthetic_config,
        )
        # Links computation
        links = await self.dataset.manager().async_links(self.dataset)
        # Update SyntheticConfig with saving_dir, Distributions,Sampling size,links
        synthetic_data_dir = os.path.join(
            self.dataset.manager().parquet_dir(), "synthetic_data"
        )
        os.makedirs(synthetic_data_dir, exist_ok=True)
        final_config = update_config(
            config_with_noise,
            synthetic_data_dir,
            parent_stats,
            links.links_statistics(),
        )

        temp_preprocess_uri = os.path.join(synthetic_data_dir, "preprocessing")
        os.makedirs(temp_preprocess_uri, exist_ok=True)
        await self.save_data_for_synthetic_training(
            final_config, temp_preprocess_uri, parent, schema
        )
        generator = SyntheticDatasetGenerator(final_config)
        dataset_schema = await dataset.manager().async_schema(dataset)
        datatype = dataset_schema.type()
        generator.train()
        sample_dict = generator.sample()
        shutil.rmtree(temp_preprocess_uri)
        sample = consolidate_samples(sample_dict, datatype, (), parent_stats)
        table = convert_array_to_table(datatype, sample)
        return async_iter_arrow(table.to_batches(max_chunksize=batch_size))

    async def save_data_for_synthetic_training(
        self,
        final_config: SyntheticConfig,
        temp_preprocess_uri: str,
        parent: st.Dataset,
        schema: st.Schema,
    ) -> None:
        # save data per table for tables that require either correlation training
        # or trigrams training or are public
        tables_with_correlation_or_trigrams_or_public: t.Dict[
            t.Tuple[str, ...], t.List[pa.RecordBatch]
        ] = {
            table_name: []
            for table_name, table_config in final_config.tables.items()
            if table_config.correlation_generation is not None
            or table_config.is_public
            or (
                table_config.independent_generation is not None
                and any(
                    isinstance(config, TrigramsColumn)
                    for config in table_config.independent_generation.columns.values()
                )
            )
        }
        if len(tables_with_correlation_or_trigrams_or_public) > 0:
            async for batch in await parent.async_to_arrow():
                array = convert_record_batch(
                    record_batch=batch, _type=schema.type()
                )
                old_arrays = array.flatten()
                idx_data = array.type.get_field_index(DATA)
                array = old_arrays[idx_data]
                for (
                    table_name
                ) in tables_with_correlation_or_trigrams_or_public.keys():
                    table_path = straight_path([DATA, *table_name])
                    updated_array, filter_indices = select_table(
                        _type=schema.data_type(),
                        array=array,
                        path=table_path,
                    )
                    old_arrays[idx_data] = updated_array
                    new_struct = pa.StructArray.from_arrays(
                        old_arrays,
                        names=list(schema.type().children().keys()),
                    )
                    tables_with_correlation_or_trigrams_or_public[
                        table_name
                    ].append(
                        pa.record_batch(new_struct.filter(filter_indices))
                    )
            for (
                table_name,
                table_list,
            ) in tables_with_correlation_or_trigrams_or_public.items():
                table = pa.Table.from_batches(table_list)
                table = table.flatten()
                table = table.rename_columns(
                    [el.removeprefix(DATA + ".") for el in table.column_names]
                )
                data_uri = os.path.join(
                    temp_preprocess_uri, str(table_name) + ".parquet"
                )
                final_config.tables[table_name].data_uri = data_uri
                pq.write_table(table, version="2.6", where=data_uri)

    async def size(self) -> st.Size:
        parent_size = await self.parent_size()
        return size_builder(self.dataset, parent_size.statistics())

    async def multiplicity(self) -> st.Multiplicity:
        parent_multiplicity = await self.parent_multiplicity()
        return multiplicity_builder(
            self.dataset, parent_multiplicity.statistics()
        )

    async def bounds(self) -> st.Bounds:
        parent_bounds = await self.parent_bounds()
        return bounds_builder(self.dataset, parent_bounds.statistics())

    async def marginals(self) -> st.Marginals:
        parent_marginals = await self.parent_marginals()
        return marginals_builder(self.dataset, parent_marginals.statistics())


class SyntheticModelStaticChecker(StandardScalarStaticChecker): ...


class SyntheticModel(StandardScalarImplementation):
    """Computes the synthetic model to use"""

    async def value(self) -> t.Any:
        attribute = self.scalar.attribute(name=SYNTHETIC_MODEL)
        assert attribute is not None, "No attribute value for synthetic model"
        return (
            RootModel[SyntheticConfig]
            .model_validate_json(attribute.properties()["config"])
            .root
        )


def consolidate_samples(
    batch_data: t.Dict[t.Tuple[str, ...], pa.Table],
    data_type: st.Type,
    curr_path: t.Tuple[str, ...],
    statistics: st.Statistics,
) -> pa.Array:
    """Visitor to put back together the sampled tables
    in the general schema (union or directly struct)"""

    class ConsolidatorVisitor(sdt.TypeVisitor):
        result = None

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            arrays = []
            child_stats = statistics.children()
            for field_name, field_type in fields.items():
                arrays.append(
                    consolidate_samples(
                        batch_data,
                        data_type=field_type,
                        curr_path=(*curr_path, field_name),
                        statistics=child_stats[field_name],
                    )
                )
            names = list(fields.keys())
            sizes = [
                getattr(
                    stat.protobuf(),
                    t.cast(str, stat.protobuf().WhichOneof("statistics")),
                ).size
                for stat in child_stats.values()
            ]
            pa_fields = [
                pa.field(name=field_name, type=array.type, nullable=True)
                for field_name, array in zip(names, arrays)
            ]
            pa_fields.append(
                pa.field(
                    name="field_selected",
                    type=pa.large_string(),
                    nullable=False,
                )
            )
            structs = []
            for i, array in enumerate(arrays):
                structs.append(
                    pa.concat_arrays(
                        [
                            pa.array([None] * sum(sizes[:i]), type=array.type),
                            array.combine_chunks()
                            if isinstance(array, pa.ChunkedArray)
                            else array,
                            pa.array(
                                sum(sizes[i + 1 :]) * [None], type=array.type
                            ),
                        ]
                    )
                )
            field_selected = pa.array(
                [
                    names[i]
                    for i, length in enumerate(sizes)
                    for _ in range(length)
                ],
                pa.large_string(),
            )
            if isinstance(field_selected, pa.ChunkedArray):
                field_selected = field_selected.combine_chunks()
            structs.append(field_selected)

            self.result = pa.StructArray.from_arrays(structs, fields=pa_fields)

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            # check that types are correct with sarus
            new_fields = [
                pa.field(
                    name=field_name,
                    type=to_arrow(fields[field_name]),
                    nullable=fields[field_name].protobuf().HasField("optional")
                    or fields[field_name].protobuf().HasField("unit"),
                )
                for field_name in fields.keys()
            ]

            self.result = pa.StructArray.from_arrays(
                [
                    el.combine_chunks()
                    for el in batch_data[tuple(curr_path)]
                    .select(list(fields.keys()))
                    .columns
                ],
                fields=new_fields,
            )

    visitor = ConsolidatorVisitor()
    data_type.accept(visitor)
    return visitor.result



from __future__ import annotations

from collections import defaultdict
import typing as t

from sarus_data_spec.base import Base
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class Path(Base[sp.Path]):
    def prototype(self) -> t.Type[sp.Path]:
        return sp.Path

    def label(self) -> str:
        return self._protobuf.label

    def sub_paths(self) -> t.List[st.Path]:
        return [Path(path) for path in self._protobuf.paths]

    def select(self, select_path: st.Path) -> t.List[st.Path]:
        assert select_path.label() == self.label()

        if len(select_path.sub_paths()) == 0:
            return self.sub_paths()

        final_sub_paths = []
        for sub_path in select_path.sub_paths():
            for available_sub_path in self.sub_paths():
                if available_sub_path.label() == sub_path.label():
                    final_sub_paths.extend(available_sub_path.select(sub_path))

        return final_sub_paths

    def is_empty(self) -> bool:
        assert not self.has_sub_paths()
        return self.label() == ""

    def has_sub_paths(self) -> bool:
        return len(self._protobuf.paths) > 0

    def to_strings_list(self) -> t.List[t.List[str]]:
        paths = []
        proto = self._protobuf
        if len(proto.paths) == 0:
            return [[proto.label]]
        for path in proto.paths:
            out = Path(path).to_strings_list()
            for el in out:
                el.insert(
                    0,
                    proto.label,
                )
            paths.extend(out)
        return paths

    def to_dict(self) -> t.Dict[str, str]:
        list_paths = self.to_strings_list()
        return {
            ".".join(path[1:-1]): path[-1] for path in list_paths
        }  # always start with DATA


def paths(path_list: t.List[t.List[str]]) -> t.List[Path]:
    out = defaultdict(list)
    for path in path_list:
        try:
            first_el = path.pop(0)
        except IndexError:
            return []
        else:
            out[first_el].append(path)
    return [
        Path(
            sp.Path(
                label=element,
                paths=[path.protobuf() for path in paths(path_list)],
            )
        )
        for element, path_list in dict(out).items()
    ]


def path(label: str = "", paths: t.Optional[t.List[st.Path]] = None) -> Path:
    """Builds a Path out of a label and a set of sub-paths"""
    if paths is None:
        paths = []
    return Path(
        sp.Path(label=label, paths=[element.protobuf() for element in paths])
    )


def straight_path(nodes: t.List[str]) -> Path:
    """Returns linear path between elements in the list"""
    if len(nodes) == 0:
        raise ValueError("At least one node must be provided")
    curr_sub_path: t.List[st.Path] = []
    for el in reversed(nodes):
        update = path(label=el, paths=curr_sub_path)
        curr_sub_path = [update]
    return update


def append_to_straight_path(
    curr_path: t.Optional[st.Path], new_element: str
) -> st.Path:
    if curr_path is None:
        return straight_path([new_element])
    else:
        return straight_path(
            [
                *(element for element in curr_path.to_strings_list()[0]),
                new_element,
            ]
        )


if t.TYPE_CHECKING:
    test_path: st.Path = Path(sp.Path())
